; Assembler code for pac-man game
; by Sam Revitch, CSE 467

include "minirisc.h"

	;; bank-shared registers
r0	equ 0x08		; Generic clobber-at-will registers
r1	equ 0x09
r2	equ 0x0a
r3	equ 0x0b
r4	equ 0x0c
r5	equ 0x0d

rngX	equ 0x0e		; Every page gets an rngX
animcnt	equ 0x0f		; and an animation counter

	;; erb0 registers
rt0	equ 0x10		; Temp register base for maze algo
				; NOTE: OVERLAP
mzcolor	equ 0x10		; Maze FG color
mzbg	equ 0x11		; Maze BG color
ptcolor	equ 0x12		; Pellet color
pmscf	equ 0x13		; Pac-man sprite config
pmcolor	equ 0x14		; Pac-man color
pmxa	equ 0x15		; Pac-man X major
pmxb	equ 0x16		; Pac-man X minor
pmya	equ 0x17		; Pac-man Y major
pmyb	equ 0x18		; Pac-man Y minor

pmdir	equ 0x19		; Pac-man's direction

frmcnt	equ 0x1b		; Frame counter
ppcntr	equ 0x1c		; power pellet countdown
pellets	equ 0x1d		; Remaining pellets
gmstate	equ 0x1e		; Game state bits
rngUC	equ 0x1f		; RNG usage count, reseeded at overflow

	;; Ghost registers
	;; NOTE these are offsets in erb1-erb3
gscf	equ 0x10		; Ghost sprite config
gcolor	equ 0x11		; Ghost color
gxa	equ 0x12		; Ghost X major
gxb	equ 0x13		; Ghost X minor
gya	equ 0x14		; Ghost Y major
gyb	equ 0x15		; Ghost Y minor

gdir	equ 0x16		; Ghost direction
grt0	equ 0x18		; Ghost temporary registers

	;; Game state bits
coin	equ 0
anim	equ 1
active	equ 2
pwrplt	equ 3
gameovr	equ 4

	;; Sprite config values
LEFT	equ 0x01
RIGHT	equ 0x00
UP	equ 0x06
DOWN	equ 0x04

	;; Keyboard scan codes for given PS2 interface
KBD_ESC		equ 0xec
KBD_SPACE	equ 0x52
KBD_UP		equ 0xea
KBD_DOWN	equ 0xe4
KBD_LEFT	equ 0xd6
KBD_RIGHT	equ 0xe8

KBD_ESC1	equ 0x76		; Controller sucks
KBD_SPACE1	equ 0x29
KBD_UP1		equ 0x75
KBD_DOWN1	equ 0x72
KBD_LEFT1	equ 0x6b
KBD_RIGHT1	equ 0x74


org 0x000

; entry point
; This gets invoked:
; (1) At hardware reset
; (2) whenever a VBLANK IRQ is received
; The hardware reset case can be distinguished by
; reading the video controller's status byte.
; This is one of the first things we do

entry	movlw 0x1f		; Enable timer, set internal
	option			; clock to on

	;; Disable interrupts to preserve internal state.
	;; We have ~400,000 instructions to execute between vblanks,
	;; which should be enough for just about every worst case.
	movlw 0xfd
	tris portc

	clrf fsr

	btfsc portc, 0		; Check hardware reset bit
	call hwreset		; Call the initializer if set

	btfss gmstate, coin
	goto inscoin

	btfss gmstate, active
	goto waiting

ingame	movf porta, 0		; Compare keycode
	movwf r0
	movlw KBD_ESC
	xorwf r0, 0
	btfsc status, z		; if equal
	goto gmreset		; goto gmreset
	movlw KBD_ESC1
	xorwf r0, 0
	btfsc status, z
	goto gmreset

	;; Do pac-man
	lcall dopman

	;; Do ghosts
	movlw erb1
	movwf fsr		; 10
	lcall ghostai
	movlw erb2
	movwf fsr
	lcall ghostai
	movlw erb3
	movwf fsr
	lcall ghostai
	clrf fsr

	;; Check collisions
	lcall gcc
	andlw 0xff		; Check result returned in W
	btfsc status, z
	goto ingplts

	movlw 0x19		; If the ghost got pacman,
	movwf pmcolor		; make him bright green

	;; Check pellet count
ingplts	movf pellets, 0
	btfsc status, z		; Reset the game when all
	goto gotcoin		; pellets are gone.

	goto frmdone


;; BEGIN seg0 FUNCTION CALLS

; clrbank - Clear one register bank
; Starts with location in FSR
; Continues until r0 == 0
clrbank	clrf indf
	incf fsr
	decfsz r0
	goto clrbank
	retlw 0

; wrcfg - Video configuration register write
; Writes a range of registers to the video
; configuration port.  Procedure is to load
; the value into portc and strobe the bit 0
; tristate pin
wrcfg	movf indf, 0
	movwf portc
	movlw 0xfc
	tris portc		; Don't forget TRIS is inverted
	movlw 0xfd		; Don't enable interrupts
	tris portc
	incf fsr, 1
	decfsz r0
	goto wrcfg
	retlw 0


; pmanim
; Update pacman sprite register based on animation counter & pmdir
; Clear FSR first!
; counter[1:0] = 00 -> scf[4:3] = 00
; counter[1:0] = 01 -> scf[4:3] = 01
; counter[1:0] = 10 -> scf[4:3] = 10
; counter[1:0] = 11 -> scf[4:3] = 01
pmanim	btfsc pmscf, 7		; pac-man halt bit
	retlw 0
	clrf pmscf
	btfsc animcnt, 0
	bsf pmscf, 3
	btfss animcnt, 1
	goto pma2
	btfss animcnt, 0
	bsf pmscf, 4

pma2	movf pmdir, 0
	iorwf pmscf, 1
	retlw 0

; Update ghost sprite register based on animation counter & gdir
; Set FSR to the correct bank first!
ganim	clrf gscf
	btfsc animcnt, 2
	bsf gscf, 3
	btfsc gdir, 2
	retlw 0
	bsf gscf, 4
	btfsc gdir, 0
	bsf gscf, 0
	retlw 0

; Clear all register banks, minus the "special"
; registers 0x00-0x07
hwreset	movlw 23
	movwf r0
	movlw r1
	movwf fsr
	call clrbank

	movlw 16
	movwf r0
	movlw erb1
	movwf fsr
	call clrbank
	decf fsr, 1
	movlw 0x0f		; Make sure ghosts are off the screen
	movwf gya

	movlw 16
	movwf r0
	movlw erb2
	movwf fsr
	call clrbank
	decf fsr, 1
	movlw 0x0f		; Make sure ghosts are off the screen
	movwf gya

	movlw 16
	movwf r0
	movlw erb3
	movwf fsr
	call clrbank
	decf fsr, 1
	movlw 0x0f
	movwf gya

	clrf fsr
	movlw 9
	movwf pmxa
	movlw 15
	movwf pmxb
	movlw 7
	movwf pmya
	movlw 0x79
	movwf pmcolor
	bsf gmstate, anim

	lcall seedrng

	retlw 0

;; Ghost Collision Check routine
;; Checks whether pacman has collided with a ghost
;; Calls gmdist manhattan distance function a few times

gcc	movf pmxa, 0
	movwf r0
	movf pmxb, 0
	movwf r1
	movf pmya, 0
	movwf r2
	movf pmyb, 0
	movwf r3
	movlw erb1
	movwf fsr
	lcall gmdist
	clrf fsr

	movf r0, 0
	btfss status, z
	goto gccg1
	movlw 7
	subwf r1, 0
	btfss status, c
	retlw 1

gccg1	movf pmxa, 0
	movwf r0
	movf pmxb, 0
	movwf r1
	movf pmya, 0
	movwf r2
	movf pmyb, 0
	movwf r3
	movlw erb2
	movwf fsr
	lcall gmdist
	clrf fsr

	movf r0, 0
	btfss status, z
	goto gccg2
	movlw 7
	subwf r1, 0
	btfss status, c
	retlw 1

gccg2	movf pmxa, 0
	movwf r0
	movf pmxb, 0
	movwf r1
	movf pmya, 0
	movwf r2
	movf pmyb, 0
	movwf r3
	movlw erb2
	movwf fsr
	lcall gmdist
	clrf fsr

	movf r0, 0
	btfss status, z
	retlw 0
	movlw 7
	subwf r1, 0
	btfss status, c
	retlw 1

	retlw 0

;; dopman - Pac-man routine
;; Evaluate keyboard input, etc.
dopman	movf pmxb, 0		; If pac-man isn't on an even cell boundary
	btfss status, z		; keep moving
	goto pmmove
	movf pmyb, 0
	btfss status, z
	goto pmmove

	bcf status, c
	rlf pmya, 0
	movwf r0
	rlf r0, 1
	rlf r0, 1
	rlf r0, 1
	movlw 2			; Find out what walls are at pacman's
	subwf pmxa, 0		; location in the maze
	iorwf r0, 0
	movwf porta		; Write to the maze address port
	nop
	nop

	movf porta, 0		; note that the write above was to the
	movwf r1		; maze memory controller, but this is KBD
	movf portb, 0
	movwf r0

	;; Check for a pellet
	andlw 0x30
	btfsc status, z
	goto pmchkup

	decf pellets, 1		; Decrease pellet count

	;; Power pellet?
	btfsc r0, 5
	goto pmpplt

	bcf r0, 4		; Clear the pellet bit
	movf r0, 0
	movwf portb

	movlw 0xc0		; Write the altered pellet bits
	tris portb
	movlw 0xff
	tris portb
	goto pmchkup

pmpplt	bcf r0, 5		; Clear the power pellet
	movf r0, 0
	movwf portb

	movlw 0xc0		; Write the altered pellet bits
	tris portb
	movlw 0xff
	tris portb

	;; TODO: power pellet function
	
pmchkup	btfss r0, 0
	goto pmchkdn
	movf r1, 0
	xorlw KBD_UP
	btfsc status, z
	goto pmmvup
	movf r1, 0
	xorlw KBD_UP1
	btfss status, z
	goto pmchkdn
pmmvup	movlw UP
	movwf pmdir
	goto pmmove

pmchkdn	btfss r0, 2
	goto pmchklt
	movf r1, 0
	xorlw KBD_DOWN
	btfsc status, z
	goto pmmvdn
	movf r1, 0
	xorlw KBD_DOWN1
	btfss status, z
	goto pmchklt
pmmvdn	movlw DOWN
	movwf pmdir
	goto pmmove

pmchklt	btfss r0, 3
	goto pmchkrt
	movf r1, 0
	xorlw KBD_LEFT
	btfsc status, z
	goto pmmvlt
	movf r1, 0
	xorlw KBD_LEFT1
	btfss status, z
	goto pmchkrt
pmmvlt	movlw LEFT
	movwf pmdir
	goto pmmove

pmchkrt	btfss r0, 1
	goto pmnodir
	movf r1, 0
	xorlw KBD_RIGHT
	btfsc status, z
	goto pmmvrt
	movf r1, 0
	xorlw KBD_RIGHT1
	btfss status, z
	goto pmnodir
pmmvrt	movlw RIGHT
	movwf pmdir
	goto pmmove

pmnodir	bsf pmscf, 7		; set the un-animated bit
	retlw 0

pmmove	bcf pmscf, 7
	movlw LEFT		; Move pac-man in the specified
	xorwf pmdir, 0		; direction.
	btfss status, z
	goto pmmovu
	decf pmxb, 1
	btfss pmxb, 5
	retlw 0
	decf pmxa, 1
	movlw 0x1f
	andwf pmxb, 1
	retlw 0

pmmovu	movlw UP		; Move pac-man in the specified
	xorwf pmdir, 0		; direction.
	btfss status, z
	goto pmmovd
	decf pmyb, 1
	btfss pmyb, 5
	retlw 0
	decf pmya, 1
	movlw 0x1f
	andwf pmyb, 1
	retlw 0

pmmovd	movlw DOWN		; Move pac-man in the specified
	xorwf pmdir, 0		; direction.
	btfss status, z
	goto pmmovr
	incf pmyb, 1
	btfss pmyb, 5
	retlw 0
	incf pmya, 1
	clrf pmyb
	retlw 0

pmmovr	incf pmxb, 1
	btfss pmxb, 5
	retlw 0
	incf pmxa, 1
	clrf pmxb
	retlw 0

; END FUNCTION CALLS


	;; Pause before starting/restarting
waiting	clrf animcnt		; Generic wait loop
	movf frmcnt, 0		; Can wait up to 4 seconds, just
	btfsc status, z		; set frmcnt appropriately
	goto ready

	goto frmdone

	;; Ready to restart
ready	btfsc gmstate, gameovr
	goto gmreset
	bsf gmstate, active
	bsf gmstate, anim
	goto frmdone

	;; Now in the "insert coin" state :-)
inscoin	
	movf porta, 0
	movwf r0
	movlw KBD_SPACE
	xorwf r0, 0
	btfsc status, z		; if equal
	goto gotcoin		; goto gotcoin
	movlw KBD_SPACE1
	xorwf r0, 0
	btfsc status, z
	goto gotcoin

	goto frmdone		; done

gmreset	call hwreset
	goto frmdone


; gotcoin - Initialize a new game
gotcoin
	lcall seedrng		; re-seed the RNG

	clrf fsr
	lcall mkmaze

	;; Initialize the game state registers
	;; E.G. move pac-man & ghosts to default locations,
	;; set ghost sprite colors, etc.

	movlw erb0
	movwf fsr

	movlw 0x03		; 00 red, 00 green, 11 blue
	movwf mzcolor		; maze foreground
	movlw 0x14		; 01 red, 01 green, 00 blue
	movwf ptcolor		; pellet color

	movlw 10		; (actually cell 8)
	movwf pmxa
	clrf pmxb
	movlw 7
	movwf pmya
	clrf pmyb
	movlw 0x79		; 11 red, 11 green, 00 blue, 1 opaque
	movwf pmcolor		; pacman primary color
	movlw LEFT
	movwf pmdir

	movlw erb1
	movwf fsr

	movlw 0x61		; 11 red, 00 green, 00 blue, 1 opaque
	movwf gcolor		; ghost0 primary color
	movlw 4
	movwf gxa
	clrf gxb
	movlw 2
	movwf gya
	clrf gyb
	movlw RIGHT
	movwf gdir
	lcall ganim		; ganim sets up sprite config

	movlw erb2
	movwf fsr
	movlw 0x6b		; 11 red, 01 green, 01 blue, 1 opaque
	movwf gcolor		; ghost1 primary color
	movlw 10
	movwf gxa
	clrf gxb
	movlw 2
	movwf gya
	clrf gyb
	movlw DOWN
	movwf gdir
	lcall ganim

	movlw erb3
	movwf fsr
	movlw 0x69		; 11 red, 01 green, 00 blue, 1 opaque
	movwf gcolor		; ghost2 primary color
	movlw 16
	movwf gxa
	clrf gxb
	movlw 2
	movwf gya
	clrf gyb
	movlw LEFT
	movwf gdir
	lcall ganim

	clrf fsr
	
	bsf gmstate, coin	; Insert a coin ;-)
	bcf gmstate, active
	bcf gmstate, anim	; temporarily not animated

	movlw 0xf0		; Init pellet counter
	movwf pellets

	movlw 0x88		; wait 120 vblanks = 2.0 secs
	movwf frmcnt		; frame counter sets wait time
	clrf animcnt		; Reset the animation counter

	goto frmdone


;; frmdone - Processing loop done
;; Step 1: write video configuration registers
;; Step 2: go to sleep and wait for VBLANK

frmdone	movlw mzcolor
	movwf fsr

	movf gmstate, 0		; Save game state
	movwf r1

	btfss r1, anim
	goto fdn1

	lcall pmanim		; WHOA THERE! lcall is a MACRO.

fdn1	movlw 9			; Colors + pacman
	movwf r0
	call wrcfg

	movlw erb1		; Ghost 0
	movwf fsr
	btfss r1, anim
	goto fdn2

	lcall ganim
fdn2	movlw 6	
	movwf r0
	call wrcfg

	movlw erb2		; Ghost 1
	movwf fsr
	btfss r1, anim
	goto fdn3
	
	lcall ganim
fdn3	movlw 6
	movwf r0
	call wrcfg

	movlw erb3		; Ghost 2
	movwf fsr
	btfss r1, anim
	goto fdn4
	
	lcall ganim
fdn4	movlw 6
	movwf r0
	call wrcfg

	clrf fsr

	incf frmcnt, 1
	movlw 0x03
	andwf frmcnt, 0
	btfsc status, z		; Animation frequency is 1/4 vblank.
	incf animcnt, 1		; This way we can time power pellet
				; events with sufficient accuracy.

	btfsc pmscf, 4		; Make extra calls to the PRNG
	goto fdn5		; every so often

	lcall rand

fdn5	
	movlw 0xff		; Re-enable interrupts
	tris portc

	;; Because Mini-RISC doesn't support SLEEP,
	;; we have to do this the hard way
sleeper goto sleeper


;;; SEGMENT 1
;;; Filled with algorithmic functions:
;;; rand, seedrng, rm3
;;; gmdist, ghostai, mkmaze

org 0x200

; Silly pseudo-random number generator
; FIX ME: better algorithm
rand	movf fsr, 0
	clrf fsr
	decf rngUC, 1
	movwf fsr
	btfss status, z
	goto rand1

; seedrng - PRNG Seed generator
; Reads the timer and XORs it a few times
; Great crypto ;-)
seedrng	movf tmr0, 0
	movwf rngX
	retlw 0

rand1	movf rngX, 0
	btfsc status, z
	comf rngX, 0
	xorlw 0xb8
	movwf rngX
	clrw
	btfsc rngX, 0
	xorlw 1
	btfsc rngX, 1
	xorlw 1
	btfsc rngX, 2
	xorlw 1
	btfsc rngX, 3
	xorlw 1
	btfsc rngX, 4
	xorlw 1
	btfsc rngX, 5
	xorlw 1
	btfsc rngX, 6
	xorlw 1
	btfsc rngX, 7
	xorlw 1
	bcf status, c
	iorwf status, 1		; Stick parity in the carry bit 30
	nop
	rlf rngX, 1		; Rotate it onto rngX
	retlw 0

; rm3 - compute rngX % 3
; NOTE: this doesn't actually do a mod, but it does constrain to 0,1,2
; Saves result in r2
rm3	clrw
	btfsc rngX, 1
	xorlw 0x01
	btfsc rngX, 3
	xorlw 0x01
	btfsc rngX, 5
	xorlw 0x01
	btfsc rngX, 7
	xorlw 0x01
	movwf r2
	btfsc rngX, 0
	incf r2, 1
	retlw 0

;; gmdist - Compute manhattan distance ghost <--> pacman
;; PMX, PMY stored in r0,r1,r2,r3
;; Results go in r0,r1
;; This is a major pain to do with an 8-bit CPU

gmdist	movf gxb, 0		; Subtract X
	subwf r1, 1
	movf gxa, 0
	btfss status, c
	incf gxa, 0
	subwf r0, 1
	movlw 0x1f
	andwf r1, 1

	movf gyb, 0		; Subtract Y
	subwf r3, 1
	movf gya, 0
	btfss status, c
	incf gya, 0
	subwf r2, 1
	movlw 0x1f
	andwf r3, 1

	btfsc r0, 5		; Negate if negative
	goto gmdnegy
	decf r0, 0
	xorlw 0xff
	movwf r0

gmdnegy	btfsc r2, 5
	goto gmdadd
	decf r2, 0
	xorlw 0xff
	movwf r2

gmdadd	movf r3, 0
	addwf r1, 1
	btfsc r1, 5
	incf r0, 1
	movlw 0x1f
	andwf r1, 1

	movf r2, 0		; r0, r1 now contain manhattan
	addwf r0, 1		; distance.

	retlw 0

; ghostai - Ghost AI routine
; decides how to move a ghost around such to capture or
; avoid pac-man.  Presently does nothing of the sort :-)
ghostai	movf gxb, 0		; If ghost not on an even cell boundary
	btfss status, z		; keep moving
	goto gmove
	movf gyb, 0
	btfss status, z
	goto gmove

	movf fsr, 0		; Copy cell location of pac-man
	movwf r0		; into r3, r4
	clrf fsr		; Make sure to back up & restore FSR, it
	movf pmxa, 0		; points to our local register space
	movwf r3
	movf pmya, 0
	movwf r4
	movf gmstate, 0		; copy game state into r5
	movwf r5		; This contains the power pellet bit. :)
	movf r0, 0
	movwf fsr

	bcf status, c
	rlf gya, 0
	movwf r0
	rlf r0, 1
	rlf r0, 1
	rlf r0, 1
	movlw 2			; Find out what walls are at the ghost's
	subwf gxa, 0		; location in the maze
	iorwf r0, 0
	movwf porta		; Write to the maze address port
	nop
	nop
	nop

	btfsc portb, 0		; Just pick the last available direction
	movlw UP		; and go.
	btfsc portb, 2
	movlw DOWN
	btfsc portb, 3		; smart AI, huh?
	movlw LEFT
	btfsc portb, 1
	movlw RIGHT
	movwf gdir

gmove	movlw LEFT		; Move the ghost in the specified
	xorwf gdir, 0		; direction.
	btfss status, z
	goto gmovr
	decf gxb, 1
	btfss gxb, 5
	retlw 0
	decf gxa, 1
	movlw 0x1f
	andwf gxb, 1
	retlw 0

gmovr	movlw RIGHT
	xorwf gdir, 0
	btfss status, z
	goto gmovd
	incf gxb, 1
	btfss gxb, 5
	retlw 0
	incf gxa
	clrf gxb
	retlw 0

gmovu	movlw UP
	xorwf gdir, 0
	btfss status, z
	goto gmovd
	decf gyb, 1
	btfss gyb, 5
	retlw 0
	decf gya, 1
	movlw 0x1f
	andwf gyb, 1
	retlw 0

gmovd	incf gyb, 1
	btfss gyb, 5
	retlw 0
	incf gya, 1
	clrf gyb
	retlw 0

; mkmaze - Maze creation algorithm
; Uses a variation of the Hunt & Kill algorithm described
; on Think Labyrinth (http://www.astrolog.org/labyrnth.htm).

;; Pass 1: Set all wall segments and set all pellet bits to 01
;; Iterates 239 times

mkmaze	movlw 0xef		; Skip the last row
	movwf r0
	movlw 0x10		; 01 pellet bits, 0000 path bits
	movwf portb

mkmp1	movf r0, 0
	movwf porta

	movlw 0xc0
	tris portb
	movlw 0xff
	tris portb

	decfsz r0
	goto mkmp1

	movf r0, 0		; Don't forget the (0,0) cell.
	movwf porta
	movlw 0xc0
	tris portb
	movlw 0xff
	tris portb
	

;; Pass 2: Hunt for carved cells with adjacent uncarved cells.
;; Carve random paths until a dead end is reached.

	movlw 0xef		; r0 holds the unvisited cell count
	movwf r0

	lcall rand

	movf rngX, 0		; Initialize it to a random number (93)
	movwf r1

	movlw 0x01
	movwf r3		; r3 is the hunt mode displacement value
				; r2 is for temps
				; rt0-rt3 get the available directions

mkmol	movf r1, 0	; Make sure we're within the row bounds of the maze af
	movlw 0xef
	subwf r1, 0		; If (r1.row >= 15)
	btfss status, c
	goto mkmhunt		; go to hunt mode

	;; Enumerate possible carving directions

	movlw rt0		; initialize the direction list
	movwf fsr

	;; Check up
	movlw 0xf0		; Skip if on top row
	andwf r1, 0
	btfsc status, z
	goto mkmcdn
	movlw 0x10		; Skip if cell above is empty
	subwf r1, 0
	movwf porta
	nop
	nop
	movlw 0x10
	subwf portb, 0
	btfss status, z
	goto mkmcdn

	movlw 0x01		; Put a 0001 on the list
	movwf indf
	incf fsr, 1

	;; Check down
mkmcdn	movlw 0xdf		; Skip if on bottom row
	subwf r1, 0
	btfss status, c
	goto mkmcrt
	movlw 0x10		; Go to cell below
	addwf r1, 0
	movwf porta		; Load contents
	nop
	nop
	movlw 0x10
	subwf portb, 0
	btfss status, z
	goto mkmcrt

	movlw 0x04		; Put a 0100 on the list
	movwf indf
	incf fsr, 1

	;; Check right
mkmcrt	incf r1, 0		; Skip if on last column
	andlw 0x0f
	btfsc status, z
	goto mkmclt
	incf r1, 0		; Go to cell to right
	movwf porta		; Load contents
	nop
	nop
	movlw 0x10
	subwf portb, 0
	btfss status, z
	goto mkmclt

	movlw 0x02		; Put a 0010 on the list
	movwf indf
	incf fsr, 1

	;; Check left
mkmclt	movlw 0x0f
	andwf r1, 0		; Skip if on first column
	btfsc status, z
	goto mkmpdir
	decf r1, 0		; Go to cell to left
	movwf porta		; Load contents
	nop
	nop
	movlw 0x10
	subwf portb, 0
	btfss status, z
	goto mkmpdir

	movlw 0x08		; Put a 1000 on the list
	movwf indf
	incf fsr, 1

	;; Look at the direction vector
mkmpdir	movlw rt0		; Determine the count (d6)
	subwf fsr, 0
	andlw 0x7f		; FSR has a stuck upper bit
	btfsc status, z		; Go to hunt mode if none are available
	goto mkmhunt
	movwf r2
	
	movlw rt0		; Reload FSR
	movwf fsr

	;; Check if one direction available
	movf r2, 0
	xorlw 0x01
	btfsc status, z
	goto mkmhdir

	;; Since we have more than one dir, we need a random number
	lcall rand

	;; What about two?
	movf r2, 0
	xorlw 0x02		; if not two
	btfss status, z		; go to the 3-4 direction handler
	goto mkmpd34

	btfsc rngX, 0
	incf fsr, 1
	goto mkmhdir

	;; Three directions?
mkmpd34	btfsc r2, 3
	goto mkmpd4

	;; Since MiniRISC lacks an easy way to do modulo..
	lcall rm3		; Mod it by 3
	movf r2, 0
	addwf fsr, 1		; Add it to FSR
	goto mkmhdir

mkmpd4	movf rngX, 0		; Random selection from four
	andlw 0x03
	addwf fsr, 1

	;; Direction decided, stored in indf
mkmhdir	movf r1, 0		; Reload maze location
	movwf porta
	decf r0, 1		; Decrement unvisited counter

	movf indf, 0
	nop
	iorwf portb, 1		; Set the bit in the maze table
	movlw 0xf0
	tris portb
	movlw 0xff
	tris portb

	btfsc indf, 0
	movlw 0xf0
	btfsc indf, 1
	movlw 0x01
	btfsc indf, 2
	movlw 0x10
	btfsc indf, 3
	movlw 0xff
	addwf r1, 1		; Add offset to the current position

	goto mkmnxt

	;; Hunt mode
	;; Search the maze for visited cells
mkmhunt	movf r3, 0		; Add the displacement factor
	addwf r1, 1

	movlw 0xef
	subwf r1, 0		; If (r1.row >= 15)
	btfss status, c
	goto mkmhunt		; skip

	movf r1, 0
	movwf porta
	nop
	nop
	movlw 0x10		; Test for an empty cell
	xorwf portb, 0
	btfsc status, z		; If empty, keep hunting
	goto mkmhunt
	;; Otherwise, fall through

mkmnxt	movf r0, 0		; Loop until all cells have been visited
	btfss status, z
	goto mkmol

; Pass 3: remove dead ends
; The creation algorithm creates perfect minimum-spanning-tree
; type mazes, which results in dead-end paths all over the maze,
; which makes pac-man impossible to play.
; This algorithm doesn't work for certain edge walls, but won't
; create any pole walls.

	movlw 0xef
	movwf r0
mkmrdl	movf r0, 0
	movwf porta
	clrf r1
	movlw 1
	nop
	btfss portb, 0		; Count walls
	addwf r1, 1
	btfss portb, 1
	addwf r1, 1
	btfss portb, 2
	addwf r1, 1
	btfss portb, 3
	addwf r1, 1

	movlw 0x03		; Are there 3?
	xorwf r1, 0
	btfsc status, z
	goto mkmrmde		; If so, remove one!

	decfsz r0, 1
	goto mkmrdl
	goto mkmout

mkmrmde	btfsc portb, 0		; If up is the exit
	bsf portb, 2		; also set down.
	btfsc portb, 1		; If right is the exit
	bsf portb, 3		; also set left.
	btfsc portb, 2		; If down is the exit
	bsf portb, 0		; also set up.
	btfsc portb, 3		; If left is the exit
	bsf portb, 1		; also set right.

	movlw 0xf0		; Write the new value
	tris portb
	movlw 0xff
	tris portb

	decfsz r0, 1		; Loop
	goto mkmrdl

mkmout

; Pass 4: place pellets
; Randomly place four power pellets

	movlw 0x04
	movwf r0		; r0 is the pellet iterator

ppltlp	lcall rand

	movf rngX, 0
	movwf r1

	movlw 0xef
	subwf r1, 0
	btfsc status, c
	goto ppltlp1

	movwf 0x10
	subwf r1, 1

ppltlp1	movf r1, 0
	movwf porta
	nop
	nop
	nop

	movf portb, 0		; This is drawn out because reading
	movwf r1		; from portb reads directly from the port,

	btfsc r1, 5		; Try again if already a power pellet
	goto ppltlp

	bcf r1, 4		; whereas writing to portb puts the
	bsf r1, 5		; result in an output register.
	movf r1, 0
	movwf portb

	movlw 0xc0
	tris portb
	movlw 0x3f
	tris portb

	decfsz r0
	goto ppltlp

	;; Clean up
	movlw 4
	movwf r0
	movlw rt0
	movwf fsr
	lcall clrbank

	retlw 0			; return
